;; Generated package description from osc.el  -*- no-byte-compile: t -*-
(define-package "osc" "0.2.0.20201201.225958" "Open Sound Control protocol library" 'nil :url "https://elpa.gnu.org/packages/osc.html" :keywords '("comm" "processes" "multimedia") :authors '(("Mario Lang" . "mlang@blind.guru")) :maintainer '("Mario Lang" . "mlang@blind.guru"))
